/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.token.saml;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Map;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Collectors;

import org.apache.commons.collections4.CollectionUtils;
import org.apache.commons.lang3.StringUtils;
import org.opensaml.core.xml.XMLObject;
import org.opensaml.core.xml.schema.XSAny;
import org.opensaml.core.xml.schema.XSBoolean;
import org.opensaml.core.xml.schema.XSInteger;
import org.opensaml.core.xml.schema.XSString;
import org.opensaml.saml.saml2.core.Assertion;
import org.opensaml.saml.saml2.core.Attribute;
import org.opensaml.saml.saml2.core.AttributeStatement;
import org.opensaml.saml.saml2.core.EncryptedAssertion;
import org.opensaml.saml.saml2.core.Response;
import org.opensaml.saml.saml2.encryption.Decrypter;
import org.opensaml.saml.security.impl.SAMLSignatureProfileValidator;
import org.opensaml.security.SecurityException;
import org.opensaml.xmlsec.encryption.support.DecryptionException;
import org.opensaml.xmlsec.signature.Signature;
import org.opensaml.xmlsec.signature.support.SignatureException;
import org.opensaml.xmlsec.signature.support.SignatureTrustEngine;

import de.ozgcloud.token.TokenAttribute;
import de.ozgcloud.token.TokenAttributes;
import de.ozgcloud.token.TokenValidationProperties.TokenValidationProperty;
import de.ozgcloud.token.common.errorhandling.TokenVerificationException;
import de.ozgcloud.token.common.errorhandling.ValidationError;
import lombok.Builder;
import net.shibboleth.utilities.java.support.resolver.CriteriaSet;

@Builder
public class SamlAttributeService {

	private final SignatureTrustEngine signatureTrustEngine;
	private final Decrypter decrypter;
	private final SAMLSignatureProfileValidator profileValidator;
	private final TokenValidationProperty tokenValidationProperty;
	private final CriteriaSet verificationCriteria;

	public TokenAttributes getAttributes(Response token) {
		validateToken(token);
		return buildTokenAttributes(decryptSamlAttributes(token), token);
	}

	void validateToken(Response token) {
		if (Objects.isNull(token.getSignature())) {
			throw new TokenVerificationException("Token signature is missing");
		}
		var validationErrors = new ArrayList<ValidationError>();
		validateSignatureProfile(token.getSignature()).ifPresent(validationErrors::add);
		validateSignature(token.getSignature()).ifPresent(validationErrors::add);
		if (CollectionUtils.isNotEmpty(validationErrors)) {
			throw new TokenVerificationException("Token validation failed", validationErrors);
		}
	}

	Optional<ValidationError> validateSignatureProfile(Signature signature) {
		try {
			profileValidator.validate(signature);
		} catch (SignatureException e) {
			return Optional.of(ValidationError.builder().message("Invalid signature profile: " + e.getMessage()).cause(e).build());
		}
		return Optional.empty();
	}

	Optional<ValidationError> validateSignature(Signature signature) {
		try {
			var isValidSignature = signatureTrustEngine.validate(signature, verificationCriteria);
			if (!isValidSignature) {
				return Optional.of(ValidationError.builder().message("Invalid token signature").build());
			}
		} catch (SecurityException e) {
			return Optional.of(ValidationError.builder().message("Error on signature validation: " + e.getMessage()).cause(e).build());
		}
		return Optional.empty();
	}

	Map<String, String> decryptSamlAttributes(Response token) {
		return token.getEncryptedAssertions().stream()
				.map(this::decryptAssertion)
				.findFirst()
				.flatMap(this::getAttributeStatement)
				.map(AttributeStatement::getAttributes).stream()
				.flatMap(Collection::stream)
				.collect(Collectors.toMap(Attribute::getName, this::getAttributeValues));
	}

	Assertion decryptAssertion(EncryptedAssertion assertion) {
		try {
			return decrypter.decrypt(assertion);
		} catch (DecryptionException ex) {
			throw new TokenVerificationException("Cannot decrypt token", ex);
		}
	}

	Optional<AttributeStatement> getAttributeStatement(Assertion assertion) {
		return assertion.getStatements().stream()
				.filter(AttributeStatement.class::isInstance)
				.map(AttributeStatement.class::cast)
				.findFirst();
	}

	String getAttributeValues(Attribute attribute) {
		return attribute.getAttributeValues().stream().map(this::getAttributeValue).collect(Collectors.joining(";"));
	}

	String getAttributeValue(XMLObject value) {
		if (Objects.isNull(value)) {
			return StringUtils.EMPTY;
		}
		return switch (value) {
			case XSString xsString -> xsString.getValue();
			case XSAny xsAny -> xsAny.getTextContent();
			case XSInteger xsInteger -> String.valueOf(xsInteger.getValue());
			case XSBoolean xsBoolean -> String.valueOf(xsBoolean.getValue());
			default -> value.toString();
		};
	}

	TokenAttributes buildTokenAttributes(Map<String, String> tokenAttributes, Response token) {
		var tokenAttributesBuilder = TokenAttributes.builder().postfachId(getPostfachId(tokenAttributes, token))
				.trustLevel(getTrustLevel(tokenAttributes));
		tokenAttributes.entrySet().stream().filter(this::isNotNamedAttribute).map(this::buildTokenAttribute)
				.forEach(tokenAttributesBuilder::otherAttribute);
		return tokenAttributesBuilder.build();
	}

	String getPostfachId(Map<String, String> tokenAttributes, Response token) {
		return tokenValidationProperty.isUseIdAsPostfachId() ? token.getID() : tokenAttributes.get(getPostfachIdKey());
	}

	String getTrustLevel(Map<String, String> tokenAttributes) {
		return tokenAttributes.get(getTrustLevelKey());
	}

	boolean isNotNamedAttribute(Map.Entry<String, String> attributeEntry) {
		return !StringUtils.equalsAny(attributeEntry.getKey(), getPostfachIdKey(), getTrustLevelKey());
	}

	String getPostfachIdKey() {
		return tokenValidationProperty.getMappings().getOrDefault(TokenAttributes.POSTFACH_ID_KEY, TokenAttributes.POSTFACH_ID_KEY);
	}

	String getTrustLevelKey() {
		return tokenValidationProperty.getMappings().getOrDefault(TokenAttributes.TRUST_LEVEL_KEY, TokenAttributes.TRUST_LEVEL_KEY);
	}

	TokenAttribute buildTokenAttribute(Map.Entry<String, String> attribute) {
		return TokenAttribute.builder().name(getMappedKey(attribute.getKey())).value(attribute.getValue()).build();
	}

	String getMappedKey(String attributeKey) {
		return tokenValidationProperty.getMappings().entrySet().stream().filter(entry -> StringUtils.equals(entry.getValue(), attributeKey))
				.map(Map.Entry::getKey).findFirst().orElse(attributeKey);
	}
}
