/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.token.saml;

import java.util.stream.Collectors;
import java.util.stream.Stream;

import org.opensaml.core.criterion.EntityIdCriterion;
import org.opensaml.core.xml.config.XMLObjectProviderRegistrySupport;
import org.opensaml.saml.criterion.ProtocolCriterion;
import org.opensaml.saml.metadata.criteria.role.impl.EvaluableProtocolRoleDescriptorCriterion;
import org.opensaml.saml.saml2.core.Response;
import org.opensaml.saml.saml2.core.impl.ResponseUnmarshaller;
import org.opensaml.saml.security.impl.SAMLSignatureProfileValidator;
import org.opensaml.security.credential.UsageType;
import org.opensaml.security.credential.criteria.impl.EvaluableEntityIDCredentialCriterion;
import org.opensaml.security.credential.criteria.impl.EvaluableUsageCredentialCriterion;
import org.opensaml.security.criteria.UsageCriterion;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.DependsOn;

import de.ozgcloud.token.TokenValidationProperties;
import de.ozgcloud.token.TokenValidationProperties.TokenValidationProperty;
import lombok.RequiredArgsConstructor;
import net.shibboleth.utilities.java.support.resolver.CriteriaSet;

@Configuration
@RequiredArgsConstructor
public class SamlConfiguration {

	private final SamlTrustEngineFactory samlTrustEngineFactory;
	private final SamlDecrypterFactory samlDecrypterFactory;

	@Bean
	ResponseUnmarshaller responseUnmarshaller() {
		return (ResponseUnmarshaller) XMLObjectProviderRegistrySupport.getUnmarshallerFactory().getUnmarshaller(Response.DEFAULT_ELEMENT_NAME);
	}

	@Bean
	@DependsOn("xmlObjectProviderRegistry")
	SamlServiceRegistry samlServiceRegistry(TokenValidationProperties tokenValidationProperties) {
		var registryBuilder = SamlServiceRegistry.builder();
		for (var tokenEntity : tokenValidationProperties.getEntities()) {
			registryBuilder.samlService(tokenEntity.getIdpEntityId(), samlAttributeService(tokenEntity));
		}
		return registryBuilder.build();
	}

	SamlAttributeService samlAttributeService(TokenValidationProperty tokenValidationProperty) {
		return SamlAttributeService.builder()
				.signatureTrustEngine(samlTrustEngineFactory.buildSamlTrustEngine(tokenValidationProperty))
				.decrypter(samlDecrypterFactory.buildDecrypter(tokenValidationProperty))
				.profileValidator(samlSignatureProfileValidator())
				.tokenValidationProperty(tokenValidationProperty)
				.verificationCriteria(buildVerificationCriteria(tokenValidationProperty.getIdpEntityId()))
				.build();
	}

	SAMLSignatureProfileValidator samlSignatureProfileValidator() {
		return new SAMLSignatureProfileValidator();
	}

	CriteriaSet buildVerificationCriteria(String idpEntityId) {
		return Stream.of(
				new EvaluableEntityIDCredentialCriterion(new EntityIdCriterion(idpEntityId)),
				new EvaluableProtocolRoleDescriptorCriterion(new ProtocolCriterion("urn:oasis:names:tc:SAML:2.0:protocol")),
				new EvaluableUsageCredentialCriterion(new UsageCriterion(UsageType.SIGNING))
		).collect(Collectors.toCollection(CriteriaSet::new));
	}

}
