/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.token.saml;

import java.io.IOException;
import java.security.cert.CertificateException;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;
import java.security.interfaces.RSAPrivateKey;
import java.util.List;

import org.opensaml.saml.saml2.encryption.Decrypter;
import org.opensaml.saml.saml2.encryption.EncryptedElementTypeEncryptedKeyResolver;
import org.opensaml.security.credential.CredentialSupport;
import org.opensaml.xmlsec.encryption.support.ChainingEncryptedKeyResolver;
import org.opensaml.xmlsec.encryption.support.EncryptedKeyResolver;
import org.opensaml.xmlsec.encryption.support.InlineEncryptedKeyResolver;
import org.opensaml.xmlsec.encryption.support.SimpleRetrievalMethodEncryptedKeyResolver;
import org.opensaml.xmlsec.keyinfo.KeyInfoCredentialResolver;
import org.opensaml.xmlsec.keyinfo.impl.CollectionKeyInfoCredentialResolver;
import org.springframework.security.converter.RsaKeyConverters;
import org.springframework.stereotype.Component;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.token.TokenValidationProperties.TokenValidationProperty;

@Component
class SamlDecrypterFactory {

	private static final String X509_CERTIFICATE_TYPE = "X.509";

	public Decrypter buildDecrypter(TokenValidationProperty tokenValidationProperty) {
		return DecrypterBuilder.builder()
				.keyEncryptionKeyResolver(buildKeyInfoCredentialResolver(tokenValidationProperty))
				.encryptedKeyElementsResolver(buildEncryptedKeyResolver())
				.build();
	}

	CollectionKeyInfoCredentialResolver buildKeyInfoCredentialResolver(TokenValidationProperty tokenValidationProperty) {
		return new CollectionKeyInfoCredentialResolver(
				List.of(CredentialSupport.getSimpleCredential(getCertificate(tokenValidationProperty), getPrivateKey(tokenValidationProperty))));
	}

	X509Certificate getCertificate(TokenValidationProperty tokenValidationProperty) {
		try (var inputStream = tokenValidationProperty.getCertificate().getInputStream()) {
			return (X509Certificate) CertificateFactory.getInstance(X509_CERTIFICATE_TYPE).generateCertificate(inputStream);
		} catch (IOException | CertificateException e) {
			throw new TechnicalException("Cannot read certificate", e);
		}
	}

	RSAPrivateKey getPrivateKey(TokenValidationProperty tokenValidationProperty) {
		try (var inputStream = tokenValidationProperty.getKey().getInputStream()) {
			return RsaKeyConverters.pkcs8().convert(inputStream);
		} catch (IOException e) {
			throw new TechnicalException("Cannot read encryption key", e);
		}
	}

	ChainingEncryptedKeyResolver buildEncryptedKeyResolver() {
		return new ChainingEncryptedKeyResolver(List.of(new InlineEncryptedKeyResolver(), new EncryptedElementTypeEncryptedKeyResolver(),
				new SimpleRetrievalMethodEncryptedKeyResolver()));
	}

	static class DecrypterBuilder {

		static DecrypterBuilder builder() {
			return new DecrypterBuilder();
		}

		private KeyInfoCredentialResolver keyEncryptionKeyResolver;
		private EncryptedKeyResolver encryptedKeyElementsResolver;

		DecrypterBuilder keyEncryptionKeyResolver(KeyInfoCredentialResolver keyEncryptionKeyResolver) {
			this.keyEncryptionKeyResolver = keyEncryptionKeyResolver;
			return this;
		}

		DecrypterBuilder encryptedKeyElementsResolver(EncryptedKeyResolver encryptedKeyElementsResolver) {
			this.encryptedKeyElementsResolver = encryptedKeyElementsResolver;
			return this;
		}

		Decrypter build() {
			var decrypter = new Decrypter(null, keyEncryptionKeyResolver, encryptedKeyElementsResolver);
			decrypter.setRootInNewDocument(true);
			return decrypter;
		}
	}
}
