/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.token.saml;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.Optional;

import org.opensaml.core.xml.io.UnmarshallingException;
import org.opensaml.saml.saml2.core.Issuer;
import org.opensaml.saml.saml2.core.Response;
import org.opensaml.saml.saml2.core.impl.ResponseUnmarshaller;
import org.springframework.stereotype.Service;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.token.TokenAttributes;
import de.ozgcloud.token.TokenValidationResult;
import de.ozgcloud.token.common.errorhandling.TokenVerificationException;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;
import net.shibboleth.utilities.java.support.xml.ParserPool;
import net.shibboleth.utilities.java.support.xml.XMLParserException;

@Service
@RequiredArgsConstructor
@Log4j2
public class SamlTokenService {

	private final SamlServiceRegistry samlServiceRegistry;
	private final ParserPool parserPool;
	private final ResponseUnmarshaller responseUnmarshaller;

	public TokenValidationResult validate(String token) {
		try {
			return buildValidTokenResult(getAttributes(parseToken(token)));
		} catch (TokenVerificationException e) {
			LOG.debug("Token validation failed", e);
			e.getValidationErrors().forEach(validationError -> LOG.error(validationError.getMessage(), validationError.getCause()));
			return buildInvalidTokenResult(e);
		}
	}

	Response parseToken(String token) {
		try (var inputStream = buildInputStream(token)) {
			var element = parserPool.parse(inputStream).getDocumentElement();
			return (Response) responseUnmarshaller.unmarshall(element);
		} catch (IOException | XMLParserException | UnmarshallingException e) {
			throw new TokenVerificationException("Cannot read token: " + e.getMessage(), e);
		}
	}

	InputStream buildInputStream(String token) {
		return new ByteArrayInputStream(token.getBytes(StandardCharsets.UTF_8));
	}

	TokenAttributes getAttributes(Response token) {
		return getSamlAttributeService(getTokenIssuer(token)).getAttributes(token);
	}

	String getTokenIssuer(Response token) {
		return Optional.ofNullable(token.getIssuer()).map(Issuer::getValue)
				.orElseThrow(() -> new TokenVerificationException("No token issuer found"));
	}

	SamlAttributeService getSamlAttributeService(String tokenIssuer) {
		return samlServiceRegistry.getService(tokenIssuer)
				.orElseThrow(() -> new TechnicalException("Can't validate token with issuer %s".formatted(tokenIssuer)));
	}

	TokenValidationResult buildValidTokenResult(TokenAttributes tokenAttributes) {
		return TokenValidationResult.builder().valid(true).attributes(tokenAttributes).build();
	}

	TokenValidationResult buildInvalidTokenResult(TokenVerificationException exception) {
		return TokenValidationResult.builder()
				.valid(false)
				.validationErrors(exception.getValidationErrors())
				.build();
	}

}
