/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */

package de.ozgcloud.token.saml;

import java.io.IOException;
import java.io.InputStream;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Objects;
import java.util.Optional;
import java.util.stream.Stream;

import org.opensaml.core.xml.XMLObject;
import org.opensaml.core.xml.config.XMLObjectProviderRegistry;
import org.opensaml.core.xml.io.Unmarshaller;
import org.opensaml.core.xml.io.UnmarshallingException;
import org.opensaml.saml.common.xml.SAMLConstants;
import org.opensaml.saml.saml2.metadata.EntitiesDescriptor;
import org.opensaml.saml.saml2.metadata.EntityDescriptor;
import org.opensaml.saml.saml2.metadata.KeyDescriptor;
import org.opensaml.security.credential.Credential;
import org.opensaml.security.credential.UsageType;
import org.opensaml.security.credential.impl.CollectionCredentialResolver;
import org.opensaml.security.x509.BasicX509Credential;
import org.opensaml.xmlsec.config.impl.DefaultSecurityConfigurationBootstrap;
import org.opensaml.xmlsec.keyinfo.KeyInfoSupport;
import org.opensaml.xmlsec.signature.support.SignatureTrustEngine;
import org.opensaml.xmlsec.signature.support.impl.ExplicitKeySignatureTrustEngine;
import org.springframework.core.io.Resource;
import org.springframework.security.saml2.core.Saml2X509Credential;
import org.springframework.stereotype.Component;
import org.w3c.dom.Element;

import de.ozgcloud.common.errorhandling.TechnicalException;
import de.ozgcloud.token.TokenValidationProperties.TokenValidationProperty;
import lombok.RequiredArgsConstructor;
import net.shibboleth.utilities.java.support.xml.ParserPool;
import net.shibboleth.utilities.java.support.xml.XMLParserException;

@Component
@RequiredArgsConstructor
class SamlTrustEngineFactory {

	private final XMLObjectProviderRegistry registry;
	private final ParserPool parserPool;

	public SignatureTrustEngine buildSamlTrustEngine(TokenValidationProperty tokenValidationProperty) {
		return new ExplicitKeySignatureTrustEngine(buildCredentialResolver(tokenValidationProperty),
				DefaultSecurityConfigurationBootstrap.buildBasicInlineKeyInfoCredentialResolver());
	}

	CollectionCredentialResolver buildCredentialResolver(TokenValidationProperty entity) {
		var credentials = getCertificatesFromMetadata(entity.getMetadata())
				.map(certificate -> buildBasicX509Credential(certificate, entity.getIdpEntityId()))
				.toList();
		if (credentials.isEmpty()) {
			throw new TechnicalException("Metadata response is missing verification certificates, necessary for verifying SAML assertions");
		}
		return new CollectionCredentialResolver(credentials);
	}

	Stream<Saml2X509Credential> getCertificatesFromMetadata(Resource metadata) {
		try (var metadataInputStream = metadata.getInputStream()) {
			return findEntityDescriptor(getXmlObject(metadataInputStream))
					.map(this::getVerificationCertificates)
					.orElseThrow(() -> new TechnicalException("No IDPSSO Descriptors found"));
		} catch (IOException e) {
			throw new TechnicalException("Error reading idp metadata.", e);
		}
	}

	XMLObject getXmlObject(InputStream inputStream) {
		try {
			var element = parserPool.parse(inputStream).getDocumentElement();
			return getUnmarshaller(element).unmarshall(element);
		} catch (XMLParserException | UnmarshallingException e) {
			throw new TechnicalException("Cannot parse metadata", e);
		}
	}

	Unmarshaller getUnmarshaller(Element element) {
		var unmarshallerFactory = registry.getUnmarshallerFactory();
		return Optional.ofNullable(unmarshallerFactory.getUnmarshaller(element))
				.orElseThrow(() -> new TechnicalException("Unsupported element of type " + element.getTagName()));
	}

	Optional<EntityDescriptor> findEntityDescriptor(XMLObject metadata) {
		return extractEntityDescriptor(metadata).filter(
				entityDescriptor -> Objects.nonNull(entityDescriptor.getIDPSSODescriptor(SAMLConstants.SAML20P_NS)));
	}

	Optional<EntityDescriptor> extractEntityDescriptor(XMLObject metadata) {
		if (metadata instanceof EntityDescriptor entityDescriptor) {
			return Optional.of(entityDescriptor);
		} else if (metadata instanceof EntitiesDescriptor entitiesDescriptor) {
			return entitiesDescriptor.getEntityDescriptors().stream().findFirst();
		}
		return Optional.empty();
	}

	Stream<Saml2X509Credential> getVerificationCertificates(EntityDescriptor descriptor) {
		var idpssoDescriptor = descriptor.getIDPSSODescriptor(SAMLConstants.SAML20P_NS);
		if (Objects.isNull(idpssoDescriptor)) {
			throw new TechnicalException("Metadata response is missing the necessary IDPSSODescriptor element");
		}
		return idpssoDescriptor.getKeyDescriptors().stream()
				.filter(this::isSignatureKey)
				.flatMap(this::getCertificates)
				.map(Saml2X509Credential::verification);
	}

	boolean isSignatureKey(KeyDescriptor keyDescriptor) {
		return keyDescriptor.getUse() == UsageType.SIGNING;
	}

	Stream<X509Certificate> getCertificates(KeyDescriptor keyDescriptor) {
		try {
			return KeyInfoSupport.getCertificates(keyDescriptor.getKeyInfo()).stream();
		} catch (CertificateException e) {
			throw new TechnicalException("Error reading certificates from key descriptor", e);
		}
	}

	Credential buildBasicX509Credential(Saml2X509Credential key, String idpEntityId) {
		var cred = new BasicX509Credential(key.getCertificate());
		cred.setUsageType(UsageType.SIGNING);
		cred.setEntityId(idpEntityId);
		return cred;
	}

}
