/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.token;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.util.Map;
import java.util.stream.Collectors;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.opensaml.xmlsec.signature.support.SignatureTrustEngine;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.test.annotation.DirtiesContext;

import de.ozgcloud.common.test.ITCase;
import de.ozgcloud.common.test.TestUtils;
import de.ozgcloud.token.saml.SamlTestConfiguration;
import lombok.SneakyThrows;
import net.devh.boot.grpc.client.inject.GrpcClient;

@ITCase
@SpringBootTest(properties = {
		"grpc.server.inProcessName=test",
		"grpc.client.token-checker.address=in-process:test",
})
@DirtiesContext
public class TokenCheckGrpcServiceITCase {

	@GrpcClient("token-checker")
	private TokenCheckServiceGrpc.TokenCheckServiceBlockingStub tokenCheckerStub;

	@MockBean
	private SignatureTrustEngine signatureTrustEngine;

	@Nested
	class TestCheckTokenSuccessfully {

		@SneakyThrows
		@BeforeEach
		void init() {
			when(signatureTrustEngine.validate(any(), any())).thenReturn(true);
		}

		@Test
		void shouldAcceptToken() {
			var token = TestUtils.loadTextFile("SamlResponseBayernId.xml");

			var result = tokenCheckerStub.checkToken(buildCheckTokenRequest(token));

			assertThat(result.getTokenValid()).isTrue();
		}

		@Test
		void shouldSetClientAttribute() {
			var token = TestUtils.loadTextFile("SamlResponseBayernId.xml");

			var result = tokenCheckerStub.checkToken(buildCheckTokenRequest(token));

			assertThat(result.getTokenAttributes().getPostfachId()).isEqualTo(SamlTestConfiguration.POSTFACH_ID);
			assertThat(result.getTokenAttributes().getTrustLevel()).isEqualTo(SamlTestConfiguration.TRUST_LEVEL);
			assertThat(getOtherFields(result)).hasSize(4).doesNotContainKeys(TokenAttributes.POSTFACH_ID_KEY, TokenAttributes.TRUST_LEVEL_KEY);
		}

		private Map<String, String> getOtherFields(GrpcCheckTokenResponse result) {
			return result.getTokenAttributes().getOtherFieldsList().stream()
					.collect(Collectors.toMap(GrpcOtherField::getName, GrpcOtherField::getValue));
		}
	}

	@Nested
	class TestCheckTokenFailure {

		@Test
		void shouldDeclineToken() {
			var token = TestUtils.loadTextFile("SamlResponseBayernId.xml");

			var result = tokenCheckerStub.checkToken(buildCheckTokenRequest(token));

			assertThat(result.getTokenValid()).isFalse();
			assertThat(result.getCheckErrors().getCheckErrorList()).hasSize(1).first().extracting(GrpcCheckError::getMessage).asString()
					.contains("Invalid token signature");
		}

	}

	private GrpcCheckTokenRequest buildCheckTokenRequest(String token) {
		return GrpcCheckTokenRequest.newBuilder().setToken(token).build();
	}
}
