/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.token.saml;

import static org.assertj.core.api.Assertions.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Random;
import java.util.UUID;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.ValueSource;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.opensaml.core.xml.XMLObject;
import org.opensaml.core.xml.schema.XSAny;
import org.opensaml.core.xml.schema.XSBoolean;
import org.opensaml.core.xml.schema.XSBooleanValue;
import org.opensaml.core.xml.schema.XSInteger;
import org.opensaml.core.xml.schema.XSString;
import org.opensaml.saml.saml2.core.Assertion;
import org.opensaml.saml.saml2.core.Attribute;
import org.opensaml.saml.saml2.core.AttributeStatement;
import org.opensaml.saml.saml2.core.EncryptedAssertion;
import org.opensaml.saml.saml2.core.Response;
import org.opensaml.saml.saml2.core.Statement;
import org.opensaml.saml.saml2.encryption.Decrypter;
import org.opensaml.saml.security.impl.SAMLSignatureProfileValidator;
import org.opensaml.security.SecurityException;
import org.opensaml.xmlsec.encryption.support.DecryptionException;
import org.opensaml.xmlsec.signature.Signature;
import org.opensaml.xmlsec.signature.support.SignatureException;
import org.opensaml.xmlsec.signature.support.SignatureTrustEngine;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.token.TokenAttributeTestFactory;
import de.ozgcloud.token.TokenAttributes;
import de.ozgcloud.token.TokenAttributesTestFactory;
import de.ozgcloud.token.TokenValidationProperties.TokenValidationProperty;
import de.ozgcloud.token.common.errorhandling.TokenVerificationException;
import de.ozgcloud.token.common.errorhandling.ValidationError;
import lombok.SneakyThrows;
import net.shibboleth.utilities.java.support.resolver.CriteriaSet;

class SamlAttributeServiceTest {

	@Spy
	@InjectMocks
	private SamlAttributeService service;

	@Mock
	private SignatureTrustEngine signatureTrustEngine;
	@Mock
	private Decrypter decrypter;
	@Mock
	private SAMLSignatureProfileValidator profileValidator;
	@Mock
	private TokenValidationProperty tokenValidationProperty;
	@Mock
	private CriteriaSet verificationCriteria;

	@Nested
	class TestGetAttributes {

		private static final Map<String, String> TOKEN_ATTRIBUTES_MAP = Map.of("key", "value");

		@Mock
		private Response token;
		@Mock
		private TokenAttributes tokenAttributes;

		@BeforeEach
		void init() {
			doNothing().when(service).validateToken(any());
			doReturn(TOKEN_ATTRIBUTES_MAP).when(service).decryptSamlAttributes(any());
			doReturn(tokenAttributes).when(service).buildTokenAttributes(any(), any());
		}

		@Test
		void shouldCallValidateToken() {
			getAttributes();

			verify(service).validateToken(token);
		}

		@Test
		void shouldCallDecryptAttributes() {
			getAttributes();

			verify(service).decryptSamlAttributes(token);
		}

		@Test
		void shouldCallBuildTokenAttributes() {
			getAttributes();

			verify(service).buildTokenAttributes(TOKEN_ATTRIBUTES_MAP, token);
		}

		@Test
		void shouldReturnResult() {
			var result = getAttributes();

			assertThat(result).isSameAs(tokenAttributes);
		}

		private TokenAttributes getAttributes() {
			return service.getAttributes(token);
		}
	}

	@Nested
	class TestValidateToken {

		@Mock
		private Response token;

		@Test
		void shouldThrowExceptionWhenMissingSignature() {
			assertThrows(TokenVerificationException.class, TestValidateToken.this::validateToken);
		}

		@Nested
		class TestValidateSuccessfully {

			@Mock
			private Signature signature;

			@BeforeEach
			void init() {
				when(token.getSignature()).thenReturn(signature);
				doReturn(Optional.empty()).when(service).validateSignatureProfile(any());
				doReturn(Optional.empty()).when(service).validateSignature(any());
			}

			@Test
			void shouldNotThrowException() {
				assertDoesNotThrow(TestValidateToken.this::validateToken);
			}

			@Test
			void shouldCallValidateSignatureProfile() {
				validateToken();

				verify(service).validateSignatureProfile(signature);
			}

			@Test
			void shouldCallValidateSignature() {
				validateToken();

				verify(service).validateSignature(signature);
			}
		}

		@Nested
		class TestValidationFailed {

			@Mock
			private Signature signature;
			@Mock
			private ValidationError signatureProfileError;
			@Mock
			private ValidationError signatureError;

			@BeforeEach
			void init() {
				when(token.getSignature()).thenReturn(signature);
				doReturn(Optional.of(signatureProfileError)).when(service).validateSignatureProfile(any());
				doReturn(Optional.of(signatureError)).when(service).validateSignature(any());
			}

			@Test
			void shouldAddInvalidSignatureProfileToException() {
				var exception = assertThrows(TokenVerificationException.class, TestValidateToken.this::validateToken);

				assertThat(exception.getValidationErrors()).contains(signatureProfileError);
			}

			@Test
			void shouldAddInvalidSignatureToException() {
				var exception = assertThrows(TokenVerificationException.class, TestValidateToken.this::validateToken);

				assertThat(exception.getValidationErrors()).contains(signatureError);
			}
		}

		private void validateToken() {
			service.validateToken(token);
		}
	}

	@Nested
	class TestValidateSignatureProfile {

		@Mock
		private Signature signature;

		@Nested
		class TestSignatureProfileValide {

			@Test
			@SneakyThrows
			void shouldCallProfileValidator() {
				service.validateSignatureProfile(signature);

				verify(profileValidator).validate(signature);
			}

			@Test
			void shouldReturnEmpty() {
				var result = service.validateSignatureProfile(signature);

				assertThat(result).isEmpty();
			}
		}

		@Nested
		class TestInvalidSignatureProfile {

			private static final String EXCEPTION_MESSAGE = LoremIpsum.getInstance().getWords(1);

			private final SignatureException signatureException = new SignatureException(EXCEPTION_MESSAGE);

			@BeforeEach
			@SneakyThrows
			void init() {
				doThrow(signatureException).when(profileValidator).validate(any());
			}

			@Test
			void shouldSetMessage() {
				var result = service.validateSignatureProfile(signature);

				assertThat(result).get().extracting(ValidationError::getMessage, STRING)
						.hasSizeGreaterThan(EXCEPTION_MESSAGE.length())
						.endsWith(EXCEPTION_MESSAGE);
			}

			@Test
			void shouldSetCause() {
				var result = service.validateSignatureProfile(signature);

				assertThat(result).get().extracting(ValidationError::getCause).isEqualTo(signatureException);
			}
		}
	}

	@Nested
	class TestValidateSignature {

		@Mock
		private Signature signature;

		@Test
		@SneakyThrows
		void shouldCallSignatureTrustEngine() {
			service.validateSignature(signature);

			verify(signatureTrustEngine).validate(signature, verificationCriteria);
		}

		@Test
		@SneakyThrows
		void shouldReturnEmpty() {
			doReturn(true).when(signatureTrustEngine).validate(any(), any());

			var result = service.validateSignature(signature);

			assertThat(result).isEmpty();
		}

		@SneakyThrows
		@Test
		void shouldSetMessageIfInvalid() {
			doReturn(false).when(signatureTrustEngine).validate(any(), any());

			var result = service.validateSignature(signature);

			assertThat(result).get().extracting(ValidationError::getMessage, STRING).isNotEmpty();
		}

		@Nested
		class TestOnException {

			private static final String EXCEPTION_MESSAGE = LoremIpsum.getInstance().getWords(1);

			private final SecurityException signatureException = new SecurityException(EXCEPTION_MESSAGE);

			@BeforeEach
			@SneakyThrows
			void init() {
				doThrow(signatureException).when(signatureTrustEngine).validate(any(), any());
			}

			@Test
			void shouldSetMessage() {
				var result = service.validateSignature(signature);

				assertThat(result).get().extracting(ValidationError::getMessage, STRING)
						.hasSizeGreaterThan(EXCEPTION_MESSAGE.length())
						.endsWith(EXCEPTION_MESSAGE);
			}

			@Test
			void shouldSetCause() {
				var result = service.validateSignature(signature);

				assertThat(result).get().extracting(ValidationError::getCause).isEqualTo(signatureException);
			}
		}
	}

	@Nested
	class TestDecryptSamlAttributes {

		@Mock
		private Response token;
		@Mock
		private EncryptedAssertion encryptedAssertion;
		@Mock
		private Assertion assertion;
		@Mock
		private AttributeStatement attributeStatement;
		@Mock
		private Attribute attribute;

		@BeforeEach
		void init() {
			doReturn(assertion).when(service).decryptAssertion(any());
			doReturn(Optional.of(attributeStatement)).when(service).getAttributeStatement(any());
			doReturn(TokenAttributeTestFactory.VALUE).when(service).getAttributeValues(attribute);
			when(token.getEncryptedAssertions()).thenReturn(List.of(encryptedAssertion));
			when(attributeStatement.getAttributes()).thenReturn(List.of(attribute));
			when(attribute.getName()).thenReturn(TokenAttributeTestFactory.NAME);
		}

		@Test
		void shouldCallDecryptAssertion() {
			decryptSamlAttributes();

			verify(service).decryptAssertion(encryptedAssertion);
		}

		@Test
		void shouldCallGetAttributeStatement() {
			decryptSamlAttributes();

			verify(service).getAttributeStatement(assertion);
		}

		@Test
		void shouldCallGetAttributeValues() {
			decryptSamlAttributes();

			verify(service).getAttributeValues(attribute);
		}

		@Test
		void shouldReturnAttributes() {
			var result = decryptSamlAttributes();

			assertThat(result).isEqualTo(TokenAttributeTestFactory.asMap());
		}

		private Map<String, String> decryptSamlAttributes() {
			return service.decryptSamlAttributes(token);
		}
	}

	@Nested
	class TestDecryptAssertion {

		@Mock
		private EncryptedAssertion encryptedAssertion;
		@Mock
		private Assertion assertion;

		@Test
		@SneakyThrows
		void shouldDecryptAssertion() {
			when(decrypter.decrypt(any(EncryptedAssertion.class))).thenReturn(assertion);

			service.decryptAssertion(encryptedAssertion);

			verify(decrypter).decrypt(encryptedAssertion);
		}

		@Test
		@SneakyThrows
		void shouldReturnResult() {
			when(decrypter.decrypt(any(EncryptedAssertion.class))).thenReturn(assertion);

			var result = service.decryptAssertion(encryptedAssertion);

			assertThat(result).isEqualTo(assertion);
		}

		@SneakyThrows
		@Test
		void shouldThrowException() {
			when(decrypter.decrypt(any(EncryptedAssertion.class))).thenThrow(DecryptionException.class);

			assertThrows(TokenVerificationException.class, () -> service.decryptAssertion(encryptedAssertion));
		}
	}

	@Nested
	class TestGetAttributeStatement {

		@Mock
		private Assertion assertion;
		@Mock
		private AttributeStatement attributeStatement;
		@Mock
		private Statement statement;

		@Test
		void shouldReturnAttributeStatement() {
			when(assertion.getStatements()).thenReturn(List.of(attributeStatement));

			var result = service.getAttributeStatement(assertion);

			assertThat(result).contains(attributeStatement);
		}

		@Test
		void shouldReturnEmptyOptional() {
			when(assertion.getStatements()).thenReturn(List.of(statement));

			var result = service.getAttributeStatement(assertion);

			assertThat(result).isEmpty();
		}
	}

	@Nested
	class TestGetAttributeValues {

		private static final String VALUE1 = LoremIpsum.getInstance().getWords(1);
		private static final String VALUE2 = LoremIpsum.getInstance().getWords(1);

		@Mock
		private Attribute attribute;
		@Mock
		private XMLObject attributeValue1;
		@Mock
		private XMLObject attributeValue2;

		@BeforeEach
		void init() {
			doReturn(VALUE1).when(service).getAttributeValue(attributeValue1);
		}

		@Test
		void shouldCallGetAttributeValue() {
			when(attribute.getAttributeValues()).thenReturn(List.of(attributeValue1));

			service.getAttributeValues(attribute);

			verify(service).getAttributeValue(attributeValue1);
		}

		@Test
		void shouldReturnAttributeValues() {
			doReturn(VALUE2).when(service).getAttributeValue(attributeValue2);
			when(attribute.getAttributeValues()).thenReturn(List.of(attributeValue1, attributeValue2));

			var result = service.getAttributeValues(attribute);

			assertThat(result).isEqualTo(VALUE1 + ";" + VALUE2);
		}
	}

	@Nested
	class TestGetAttributeValue {

		@Test
		void shouldReturnEmptyString() {
			var result = service.getAttributeValue(null);

			assertThat(result).isEmpty();
		}

		@Test
		void shouldReturnWhenXSString() {
			XMLObject attrValue = when(mock(XSString.class).getValue()).thenReturn(TokenAttributeTestFactory.VALUE).getMock();

			var result = service.getAttributeValue(attrValue);

			assertThat(result).isEqualTo(TokenAttributeTestFactory.VALUE);
		}

		@Test
		void shouldReturnWhenXSAny() {
			XMLObject attrValue = when(mock(XSAny.class).getTextContent()).thenReturn(TokenAttributeTestFactory.VALUE).getMock();

			var result = service.getAttributeValue(attrValue);

			assertThat(result).isEqualTo(TokenAttributeTestFactory.VALUE);
		}

		@Test
		void shouldReturnWhenXSInteger() {
			var value = new Random().nextInt();
			XMLObject attrValue = when(mock(XSInteger.class).getValue()).thenReturn(value).getMock();

			var result = service.getAttributeValue(attrValue);

			assertThat(result).isEqualTo(String.valueOf(value));
		}

		@DisplayName("should return value of XSBoolean when")
		@ParameterizedTest(name = "value is {0}")
		@ValueSource(strings = { "true", "false" })
		void shouldReturnWhenXSBoolean(String value) {
			XMLObject attrValue = when(mock(XSBoolean.class).getValue()).thenReturn(XSBooleanValue.valueOf(value)).getMock();

			var result = service.getAttributeValue(attrValue);

			assertThat(result).isEqualTo(value);
		}
	}

	@Nested
	class TestBuildTokenAttributes {

		private static final Map<String, String> TOKEN_ATTRIBUTES_MAP = TokenAttributeTestFactory.asMap();

		@Mock
		private Response token;

		@BeforeEach
		void init() {
			doReturn(true).when(service).isNotNamedAttribute(any());
			doReturn(TokenAttributesTestFactory.OTHER_ATTRIBUTE).when(service).buildTokenAttribute(any());
		}

		@Test
		void shouldCallGetPostfachId() {
			buildTokenAttributes();

			verify(service).getPostfachId(TOKEN_ATTRIBUTES_MAP, token);
		}

		@Test
		void shouldSetPostfachId() {
			doReturn(TokenAttributesTestFactory.POSTFACH_ID).when(service).getPostfachId(any(), any());

			var result = buildTokenAttributes();

			assertThat(result.getPostfachId()).isEqualTo(TokenAttributesTestFactory.POSTFACH_ID);
		}

		@Test
		void shouldCallGetTrustLevel() {
			buildTokenAttributes();

			verify(service).getTrustLevel(TOKEN_ATTRIBUTES_MAP);
		}

		@Test
		void shouldSetTrustLevel() {
			doReturn(TokenAttributesTestFactory.TRUST_LEVEL).when(service).getTrustLevel(any());

			var result = buildTokenAttributes();

			assertThat(result.getTrustLevel()).isEqualTo(TokenAttributesTestFactory.TRUST_LEVEL);
		}

		@Test
		void shouldCallIsNotMappedField() {
			buildTokenAttributes();

			verify(service).isNotNamedAttribute(Map.entry(TokenAttributeTestFactory.NAME, TokenAttributeTestFactory.VALUE));
		}

		@Test
		void shouldCallBuildTokenAttribute() {
			buildTokenAttributes();

			verify(service).buildTokenAttribute(Map.entry(TokenAttributeTestFactory.NAME, TokenAttributeTestFactory.VALUE));
		}

		@Test
		void shouldSetOtherAttributes() {
			var result = buildTokenAttributes();

			assertThat(result.getOtherAttributes()).containsExactly(TokenAttributesTestFactory.OTHER_ATTRIBUTE);
		}

		private TokenAttributes buildTokenAttributes() {
			return service.buildTokenAttributes(TOKEN_ATTRIBUTES_MAP, token);
		}
	}

	@Nested
	class TestGetPostfachId {

		@Mock
		private Response token;

		@Nested
		class TestTokenId {

			private static final String TOKEN_ID = UUID.randomUUID().toString();

			@BeforeEach
			void init() {
				when(tokenValidationProperty.isUseIdAsPostfachId()).thenReturn(true);
				doReturn(TOKEN_ID).when(token).getID();
			}

			@Test
			void shouldCallGetTokenId() {
				getPostfachId();

				verify(service).getPostfachId(TokenAttributeTestFactory.asMap(), token);
			}

			@Test
			void shouldReturnTokenId() {
				var result = getPostfachId();

				assertThat(result).contains(TOKEN_ID);
			}
		}

		@Nested
		class TestTokenAttribute {

			@BeforeEach
			void init() {
				when(tokenValidationProperty.isUseIdAsPostfachId()).thenReturn(false);
				doReturn(TokenAttributeTestFactory.NAME).when(service).getPostfachIdKey();
			}

			@Test
			void shouldCallGetMappedValue() {
				getPostfachId();

				verify(service).getPostfachIdKey();
			}

			@Test
			void shouldReturnMappedValue() {
				var result = getPostfachId();

				assertThat(result).contains(TokenAttributeTestFactory.VALUE);
			}
		}

		private String getPostfachId() {
			return service.getPostfachId(TokenAttributeTestFactory.asMap(), token);
		}
	}

	@Nested
	class TestGetTrustLevel {

		@BeforeEach
		void init() {
			doReturn(TokenAttributeTestFactory.NAME).when(service).getTrustLevelKey();
		}

		@Test
		void shouldCallGetMappedValue() {
			getTrustLevel();

			verify(service).getTrustLevelKey();
		}

		@Test
		void shouldReturnMappedValue() {
			var result = getTrustLevel();

			assertThat(result).contains(TokenAttributeTestFactory.VALUE);
		}

		private String getTrustLevel() {
			return service.getTrustLevel(TokenAttributeTestFactory.asMap());
		}
	}

	@Nested
	class TestIsNotNamedAttribute {

		private static final String KEY = UUID.randomUUID().toString();

		@Test
		void shouldReturnTrueWhenNotMapped() {
			var random = new Random();
			doReturn("postfachIdKey-" + random.nextInt()).when(service).getPostfachIdKey();
			doReturn("trustLevelKey-" + random.nextInt()).when(service).getTrustLevelKey();

			var result = isNotMappedField();

			assertThat(result).isTrue();
		}

		@Test
		void shouldReturnFalseWhenMapped() {
			doReturn(KEY).when(service).getPostfachIdKey();

			var result = isNotMappedField();

			assertThat(result).isFalse();
		}

		private boolean isNotMappedField() {
			return service.isNotNamedAttribute(Map.entry(KEY, TokenAttributeTestFactory.VALUE));
		}
	}

	@Nested
	class TestGetPostfachIdKey {

		@Test
		void shouldReturnDefaultKey() {
			var result = service.getPostfachIdKey();

			assertThat(result).isEqualTo(TokenAttributes.POSTFACH_ID_KEY);
		}

		@Test
		void shouldReturnMappedKey() {
			when(tokenValidationProperty.getMappings()).thenReturn(Map.of(TokenAttributes.POSTFACH_ID_KEY, TokenAttributeTestFactory.NAME));

			var result = service.getPostfachIdKey();

			assertThat(result).isEqualTo(TokenAttributeTestFactory.NAME);
		}
	}

	@Nested
	class TestGetTrustLevelKey {

		@Test
		void shouldReturnDefaultKey() {
			var result = service.getTrustLevelKey();

			assertThat(result).isEqualTo(TokenAttributes.TRUST_LEVEL_KEY);
		}

		@Test
		void shouldReturnMappedKey() {
			when(tokenValidationProperty.getMappings()).thenReturn(Map.of(TokenAttributes.TRUST_LEVEL_KEY, TokenAttributeTestFactory.NAME));

			var result = service.getTrustLevelKey();

			assertThat(result).isEqualTo(TokenAttributeTestFactory.NAME);
		}
	}

	@Nested
	class TestBuildTokenAttribute {

		@Test
		void shouldBuildTokenAttribute() {
			var result = service.buildTokenAttribute(Map.entry(TokenAttributeTestFactory.NAME, TokenAttributeTestFactory.VALUE));

			assertThat(result).usingRecursiveComparison().isEqualTo(TokenAttributeTestFactory.create());
		}
	}

	@Nested
	class TestGetMappedKey {

		@Test
		void shouldReturnDefaultKey() {
			var result = service.getMappedKey(TokenAttributeTestFactory.VALUE);

			assertThat(result).isEqualTo(TokenAttributeTestFactory.VALUE);
		}

		@Test
		void shouldReturnMappedKey() {
			when(tokenValidationProperty.getMappings()).thenReturn(Map.of(TokenAttributeTestFactory.NAME, TokenAttributeTestFactory.VALUE));

			var result = service.getMappedKey(TokenAttributeTestFactory.VALUE);

			assertThat(result).isEqualTo(TokenAttributeTestFactory.NAME);
		}
	}
}