/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.token.saml;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.Mockito.*;

import java.util.List;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.InjectMocks;
import org.mockito.Mock;
import org.mockito.Spy;
import org.opensaml.saml.saml2.encryption.Decrypter;
import org.opensaml.saml.security.impl.SAMLSignatureProfileValidator;
import org.opensaml.xmlsec.signature.support.SignatureTrustEngine;

import com.thedeanda.lorem.LoremIpsum;

import de.ozgcloud.token.TokenValidationProperties;
import de.ozgcloud.token.TokenValidationProperties.TokenValidationProperty;
import net.shibboleth.utilities.java.support.resolver.CriteriaSet;

class SamlConfigurationTest {

	private static final String IDP_ENTITY_ID = LoremIpsum.getInstance().getWords(1);

	@Spy
	@InjectMocks
	private SamlConfiguration configuration;

	@Mock
	private SamlTrustEngineFactory samlTrustEngineFactory;
	@Mock
	private SamlDecrypterFactory samlDecrypterFactory;

	@Nested
	class TestSamlServiceRegistry {

		@Mock
		private TokenValidationProperties tokenValidationProperties;
		@Mock
		private TokenValidationProperty tokenValidationProperty;
		@Mock
		private SamlAttributeService tokenValidationService;

		@BeforeEach
		void init() {
			when(tokenValidationProperty.getIdpEntityId()).thenReturn(IDP_ENTITY_ID);
			when(tokenValidationProperties.getEntities()).thenReturn(List.of(tokenValidationProperty));
			doReturn(tokenValidationService).when(configuration).samlAttributeService(any());
		}

		@Test
		void shouldCallSamlTokenService() {
			configuration.samlServiceRegistry(tokenValidationProperties);

			verify(configuration).samlAttributeService(tokenValidationProperty);
		}

		@Test
		void shouldAddService() {
			var result = configuration.samlServiceRegistry(tokenValidationProperties);

			assertThat(result.getService(IDP_ENTITY_ID)).contains(tokenValidationService);
		}
	}

	@Nested
	class TestSamlTokenService {

		@Mock
		private TokenValidationProperty tokenValidationProperty;
		@Mock
		private SignatureTrustEngine signatureTrustEngine;
		@Mock
		private Decrypter decrypter;
		@Mock
		private SAMLSignatureProfileValidator profileValidator;
		@Mock
		private CriteriaSet verificationCriteria;

		@BeforeEach
		void init() {
			when(tokenValidationProperty.getIdpEntityId()).thenReturn(IDP_ENTITY_ID);
			doReturn(signatureTrustEngine).when(samlTrustEngineFactory).buildSamlTrustEngine(any());
			doReturn(decrypter).when(samlDecrypterFactory).buildDecrypter(any());
			doReturn(profileValidator).when(configuration).samlSignatureProfileValidator();
			doReturn(verificationCriteria).when(configuration).buildVerificationCriteria(any());
		}

		@Test
		void shouldCallBuildSamlTrustEngine() {
			samlTokenService();

			verify(samlTrustEngineFactory).buildSamlTrustEngine(tokenValidationProperty);
		}

		@Test
		void shouldSetSignatureTrustEngine() {
			var result = samlTokenService();

			assertThat(result).extracting("signatureTrustEngine").isEqualTo(signatureTrustEngine);
		}

		@Test
		void shouldCallBuildDecrypter() {
			samlTokenService();

			verify(samlDecrypterFactory).buildDecrypter(tokenValidationProperty);
		}

		@Test
		void shouldSetDecrypter() {
			var result = samlTokenService();

			assertThat(result).extracting("decrypter").isEqualTo(decrypter);
		}

		@Test
		void shouldCallSamlSignatureProfileValidator() {
			samlTokenService();

			verify(configuration).samlSignatureProfileValidator();
		}

		@Test
		void shouldSetProfileValidator() {
			var result = samlTokenService();

			assertThat(result).extracting("profileValidator").isEqualTo(profileValidator);
		}

		@Test
		void shouldSetTokenValidationProperty() {
			var result = samlTokenService();

			assertThat(result).extracting("tokenValidationProperty").isEqualTo(tokenValidationProperty);
		}

		@Test
		void shouldCallBuildVerificationCriteria() {
			samlTokenService();

			verify(configuration).buildVerificationCriteria(IDP_ENTITY_ID);
		}

		@Test
		void shouldSetVerificationCriteria() {
			var result = samlTokenService();

			assertThat(result).extracting("verificationCriteria").isEqualTo(verificationCriteria);
		}

		private SamlAttributeService samlTokenService() {
			return configuration.samlAttributeService(tokenValidationProperty);
		}
	}
}