/*
 * Copyright (C) 2024 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.ozgcloud.token.saml;

import static org.mockito.Mockito.*;

import java.util.List;
import java.util.Map;
import java.util.UUID;

import org.opensaml.core.xml.schema.XSString;
import org.opensaml.saml.saml2.core.Assertion;
import org.opensaml.saml.saml2.core.Attribute;
import org.opensaml.saml.saml2.core.AttributeStatement;
import org.opensaml.saml.saml2.core.EncryptedAssertion;
import org.opensaml.saml.saml2.encryption.Decrypter;
import org.opensaml.xmlsec.signature.support.SignatureTrustEngine;
import org.springframework.context.annotation.Bean;
import org.springframework.context.annotation.Configuration;
import org.springframework.context.annotation.Primary;

import lombok.SneakyThrows;

@Configuration
public class SamlTestConfiguration {

	public static final String POSTFACH_ID = UUID.randomUUID().toString();
	public static final String TRUST_LEVEL = "STORK-QAA-Level-1";

	public static final Map<String, String> SAML_ATTRIBUTES_MAP = Map.of(
			"urn:oid:1.2.40.0.10.2.1.1.261.94", TRUST_LEVEL,
			"urn:oid:2.5.4.18", POSTFACH_ID,
			"urn:oid:0.9.2342.19200300.100.1.3", "mail@mail.local",
			"urn:oid:1.3.6.1.4.1.25484.494450.2", "Benutzername",
			"urn:oid:2.5.4.42", "Ozg",
			"urn:oid:2.5.4.4", "Mgm"
	);

	@SneakyThrows
	@Bean
	@Primary
	SamlTrustEngineFactory samlTrustEngineFactory(SignatureTrustEngine signatureTrustEngine) {
		return when(mock(SamlTrustEngineFactory.class).buildSamlTrustEngine(any())).thenReturn(signatureTrustEngine).getMock();
	}

	@SneakyThrows
	@Bean
	@Primary
	SamlDecrypterFactory samlDecrypterFactory() {
		var attributes = SAML_ATTRIBUTES_MAP.entrySet().stream().map(e -> initAttributeMock(e.getKey(), e.getValue())).toList();
		AttributeStatement attributeStatement = when(mock(AttributeStatement.class).getAttributes()).thenReturn(attributes).getMock();
		Assertion assertion = when(mock(Assertion.class).getStatements()).thenReturn(List.of(attributeStatement)).getMock();
		Decrypter decrypter = when(mock(Decrypter.class).decrypt(any(EncryptedAssertion.class))).thenReturn(assertion).getMock();
		return when(mock(SamlDecrypterFactory.class).buildDecrypter(any())).thenReturn(decrypter).getMock();
	}

	Attribute initAttributeMock(String name, String value) {
		Attribute attribute = mock(Attribute.class);
		when(attribute.getName()).thenReturn(name);
		XSString attrValue = when(mock(XSString.class).getValue()).thenReturn(value).getMock();
		when(attribute.getAttributeValues()).thenReturn(List.of(attrValue));
		return attribute;
	}

}